.framed {
  border-radius: 5px;
  background-color: #F9F9F9;
  margin: 10px;
  padding: 15px;
  position: relative;
  box-shadow: 2px 2px 2px lightgrey;
}

.framed:hover {
  box-shadow: 5px 5px 5px 3px lightgrey;
}
